! ******************************************************************
! ******************************************************************
subroutine algencan(fsub,gsub,hsub,csub,jacsub,hcsub,fcsub,gjacsub, &
     gjacpsub,hlsub,hlpsub,jcnnzmax,hnnzmax,epsfeas,epsopt,efstain, &
     eostain,efacc,eoacc,outputfnm,specfnm,nvparam,vparam,norig, &
     xorig,lorig,uorig,m,lambda,equatn,linear,coded,checkder,fu, &
     cnormu,snorm,nlpsupn,inform)

  use modmachconst, only: inimachconst,bignum,macheps,macheps12,macheps13,macheps23
  use modouttyp,    only: iniouttyp,iprint,iprintctl,iprintinn,iprintout,mprint,ncomp,nprint
  use modalgparam,  only: hptype,precond,innslvr,innercall,ignoref,lsssubACC,lsssubNW, &
       lsssubTR,sclsubACC,sclsubNW,sclsubTR,skipacc,useustp,maxoutitgiven,&
       maxinnitgiven,maxaccitgiven,rhoinigiven,rhomaxgiven
  use modalgconst,  only: n0,n1,n2

  use lssma57
  use lssma86
  use lssma97
  use problvls, only: sinip,sendp,scale
  use problvlt, only: tinip,tendp,slacks
  use problvlu, only: uinip,uendp,rmfixv
  use problvlv, only: vinip,vendp,reperr,fcnt,safemode
  use probgiven, only: iniprobgiven,firstde,seconde,truehpr,gjacpcoded

  implicit none

  ! SCALAR ARGUMENTS
  logical,      intent(in)    :: checkder
  integer,      intent(in)    :: hnnzmax,jcnnzmax,m,norig,nvparam
  integer,      intent(out)   :: inform
  real(kind=8), intent(inout) :: cnormu,efacc,efstain,eoacc,eostain, &
                                 epsfeas,epsopt,fu,nlpsupn,snorm

  ! ARRAY ARGUMENTS
  character(len=80), intent(in)    :: specfnm,outputfnm,vparam(nvparam)
  logical,           intent(in)    :: coded(11),linear(m)
  logical,           intent(inout) :: equatn(m)
  real(kind=8),      intent(in)    :: lorig(norig),uorig(norig)
  real(kind=8),      intent(inout) :: lambda(m),xorig(norig)

  ! EXTERNAL SUBROUTINES ARGUMENTS
  external :: fsub,gsub,hsub,csub,jacsub,hcsub,fcsub,gjacsub,gjacpsub, &
              hlsub,hlpsub

  ! LOCAL SCALARS
  logical      :: innfail,lss,scl
  integer      :: alinfo,allocerr,geninfo,iter,maxit,n,nwcalls,nwtotit, &
                  outiter,solinfo,totiter,msqcalls,msqtotit
  real(kind=8) :: cnorm,cnormb,cnormub,f,fb,fub,ncsupn,nlpsupnb,rsupn
  real(kind=4) :: time

  ! LOCAL ARRAYS
  character(len=15)     :: subname
  real(kind=8)          :: rho(m)
  real(kind=8), pointer :: l(:),u(:),x(:)
  real(kind=4)          :: tdum(2)

  ! DATA STATEMENTS
  data tdum/0.0,0.0/

  ! ==================================================================
  ! Start timing
  ! ==================================================================

  time = dtime(tdum)

  ! ==================================================================
  ! Error tracker
  ! ==================================================================

  inform = 0

  ! ==================================================================
  ! Initialization
  ! ==================================================================

  ! Set machine-dependent constants (module machconst)
  call inimachconst()

  ! Output control (module outtyp)
  call iniouttyp()

  ! ==================================================================
  ! Set default values in 'prob' modules (probgiven, problvl[s|t|u|v])
  ! ==================================================================

  ! Initiate module probgiven
  call iniprobgiven(fsub,gsub,hsub,csub,jacsub,hcsub,fcsub,gjacsub, &
       gjacpsub,hlsub,hlpsub,coded,m,jcnnzmax,hnnzmax,inform)
  if ( inform .lt. 0 ) return

  ! Set default values for variables in module problvlv
  safemode = .false.

  ! Set default values for variables in module problvlu
  rmfixv = .true.

  ! Set default values for variables in module problvlt
  slacks = .false.

  ! Set default values for variables in module problvls
  scale = .true.

  ! ==================================================================
  ! Set default values for algoritmic parameters (module algparam)
  ! ==================================================================

  ! Hessian-vector product strategy: HAPPRO, INCQUO or TRUEHP. TRUEHP
  ! is the default option. If the proper subroutines were not coded by
  ! the user, then HAPPRO is used instead. HAPPRO reduces to INCQUO in
  ! the unconstrained and bound-constrained cases and switches between
  ! INCQUO and the product by an approximation of the Hessian in the
  ! constrained case.

  if ( truehpr ) then
     hptype = 'TRUEHP'
  else if ( .not. gjacpcoded ) then
     hptype = 'HAPPRO'
  else
     hptype = 'INCQUO'
  end if

  ! Conjugate Gradients preconditioner

  if ( .not. gjacpcoded ) then
     precond = 'QNCGNA'
  else
     precond = 'NONE'
  end if

  ! Subroutines for Linear systems

  if ( lss_ma57() ) then
     lsssubTR = 'MA57'
     sclsubTR = 'NONE'
  else
     lsssubTR = 'NONE'
     sclsubTR = 'NONE'
  end if

  if ( lss_ma57() ) then
     lsssubNW = 'MA57'
     sclsubNW = 'NONE'
  else if ( lss_ma97() ) then
     lsssubNW = 'MA97'
     sclsubNW = 'NONE'
  else if ( lss_ma86() ) then
     lsssubNW = 'MA86'
     sclsubNW = 'NONE'
  else
     lsssubNW = 'NONE'
     sclsubNW = 'NONE'
  end if

  lsssubACC = lsssubNW
  sclsubACC = sclsubNW

  ! Inner solver method: TR (trust regions, i.e. BETRA), NW (Newtonian
  ! system with the use of a direct solver, i.e. MA57AD), TN + TRUEHP
  ! (Truncated Newton and true Hessian-vector product) or TN + HAPPRO
  ! (Truncated Newton and product with Hessian approximation). If the
  ! proper subroutines were not provided by the user then TN + HAPPRO
  ! is used.

  if ( seconde .and. lsssubTR .ne. 'NONE' .and. norig .le. n0 ) then
     innslvr = 'TR'
  else if ( seconde .and. lsssubNW .ne. 'NONE' .and. norig .le. n1 ) then
     innslvr = 'NW'
  else
     innslvr = 'TN'
!!$     if ( norig .gt. n2 ) then
!!$        hptype  = 'HAPPRO'
!!$     end if
  end if

  ! Ignore objective function (to only find a feasible point by
  ! minimizing 1/2 of the squared infeasibility)
  ignoref = .false.

  ! Acceleration step

  if ( seconde .and. lsssubACC .ne. 'NONE' ) then
     skipacc = .false.
  else
     skipacc = .true.
  end if

  ! Maximum number of outer iterations, inner iterations, and initial
  ! penalty parameter

  maxaccitgiven = - 1
  maxinnitgiven = - 1
  maxoutitgiven = - 1
  rhoinigiven   = - 1.0d0
  rhomaxgiven   = - 1.0d0

  ! Flags used to make inner calls to Gencan

  innercall = .false.
  useustp   = .false.

  ! ==================================================================
  ! Open output file
  ! ==================================================================

  if ( outputfnm .ne. '' ) then
     open(unit=10,file=outputfnm,status='replace')
  else
     open(unit=10,               status='scratch')
  end if

  ! ==================================================================
  ! Modifiy default values using specification file
  ! ==================================================================

  call procpar(epsfeas,epsopt,efstain,eostain,efacc,eoacc,outputfnm, &
       specfnm,nvparam,vparam)

  ! ==================================================================
  ! Initialize problem data structures
  ! ==================================================================

  allocate(x(norig),l(norig),u(norig),stat=allocerr)

  if ( allocerr .ne. 0 ) then
     inform = - 93
     subname = 'ALGENCAN'
     call reperr(inform,subname)
     return
  end if

  n = norig

  x(1:norig) = xorig(1:norig)
  l(1:norig) = lorig(1:norig)
  u(1:norig) = uorig(1:norig)

  ! Initiate module problvlv
  call vinip(n,x,l,u,m,lambda,equatn,linear,coded,checkder,inform)
  if ( inform .lt. 0 ) return

  ! Initiate module problvlu
  call uinip(n,x,l,u,m,lambda,equatn,linear,coded,inform)
  if ( inform .lt. 0 ) return

  ! Initiate module problvlt
  call tinip(n,x,l,u,m,lambda,equatn,linear,coded,inform)
  if ( inform .lt. 0 ) return

  ! Initiate module problvls
  call sinip(n,x,l,u,m,lambda,equatn,linear,coded,inform)
  if ( inform .lt. 0 ) return

  nprint = min( n, ncomp )
  mprint = min( m, ncomp )

  ! ==================================================================
  ! Allocate arrays modules svdgrad, svdhess, and minsq
  ! ==================================================================

  call allocatearrays(n,m,jcnnzmax,hnnzmax,allocerr)

  if ( allocerr .ne. 0 ) then
     inform = - 93
     subname = 'ALGENCAN'
     call reperr(inform,subname)
     return
  end if

  ! ==================================================================
  ! Call the solver
  ! ==================================================================

  if ( .not. ignoref .and. m .gt. 0 ) then

     ! ALGENCAN for PNL problems

     call auglag(n,x,l,u,m,lambda,equatn,linear,epsfeas,epsopt, &
          efstain,eostain,efacc,eoacc,f,cnorm,snorm,nlpsupn,fu, &
          cnormu,fub,cnormub,fb,cnormb,nlpsupnb,ncsupn,rsupn, &
          outiter,totiter,nwcalls,nwtotit,msqcalls,msqtotit,innfail, &
          alinfo,inform)

     solinfo = alinfo

  else
     ! GENCAN for box-constrained problems and feasibility problems

     maxit = 999999999

     if ( maxinnitgiven .gt. 0 ) then
        maxit = maxinnitgiven
     end if

     ! Used in feasibility problems (ignoref=true). With lambda=0 and
     ! rho=1, to minimize 1/2 of the squared infeasibility coincides
     ! with minimizing the augmented Lagrangian.

     lambda(1:m) = 0.0d0
     rho(1:m) = 1.0d0

     call gencan(n,x,l,u,m,lambda,equatn,linear,rho,epsfeas,epsopt, &
          efstain,eostain,maxit,iter,f,nlpsupn,cnorm,cnormu,geninfo, &
          inform)

     solinfo  = geninfo

     ncsupn   = 0.0d0
     rsupn    = 0.0d0

     outiter  = 0
     totiter  = iter
     nwcalls  = 0
     nwtotit  = 0
     msqcalls = 0
     msqtotit = 0
     innfail  = .false.

     if ( ignoref ) then
        f = 0.0d0
     end if

     fu = f

     fb       = f
     fub      = fu
     cnormb   = cnorm
     cnormub  = cnormu
     nlpsupnb = nlpsupn
  end if

  if ( inform .lt. 0 ) return

  ! ==================================================================
  ! End problem data structures
  ! ==================================================================

  ! Unscale Lagrange multipliers
  call sendp(n,x,l,u,m,lambda,equatn,linear,inform)
  if ( inform .lt. 0 ) return

  ! Remove slack variables  (x, l and u may be resized)
  call tendp(n,x,l,u,m,lambda,equatn,linear,inform)
  if ( inform .lt. 0 ) return

  ! Restore fixed variables (x, l and u may be resized)
  call uendp(n,x,l,u,m,lambda,equatn,linear,inform)
  if ( inform .lt. 0 ) return

  ! Save solution (primal and dual variables)
  call vendp(n,x,l,u,m,lambda,equatn,linear,inform)
  if ( inform .lt. 0 ) return

  xorig(1:n) = x(1:n)

  ! ==================================================================
  ! Stop timing
  ! ==================================================================

  time = dtime(tdum)
  time = tdum(1)

  if ( iprintctl(4) ) then
     write(* ,9000) time
     write(10,9000) time
  end if

  ! ==================================================================
  ! Close output file
  ! ==================================================================

  close(10)

  ! ==================================================================
  ! Write statistics file with table line
  ! ==================================================================

  if ( iprintctl(5) ) then
     open(20,file='algencan-tabline.out')
     write(20,9010) fu,cnormu,f,cnorm,nlpsupn,fub,cnormub,fb,cnormb, &
                    nlpsupnb,ncsupn,rsupn,inform,solinfo,innfail,n,  &
                    m,outiter,totiter,fcnt,nwcalls,nwtotit,msqcalls, &
                    msqtotit,time
     close(20)
  end if

  ! ==================================================================
  ! Deallocate global arrays in modules
  ! ==================================================================

  deallocate(x,l,u,stat=allocerr)
  if ( allocerr .ne. 0 ) then
     inform = - 94
     subname = 'ALGENCAN'
     call reperr(inform,subname)
     return
  end if

  call deallocatearrays(allocerr)
  if ( allocerr .ne. 0 ) then
     inform = - 94
     subname = 'ALGENCAN'
     call reperr(inform,subname)
     return
  end if

  ! ==================================================================
  ! NON-EXECUTABLE STATEMENTS
  ! ==================================================================

 9000 format(/,1X,'Total CPU time in seconds: ',F8.2)

 9010 format(1X,1P,D24.16,1X,1P,D7.1,1X,1P,D24.16,1X,1P,D7.1,1X,1P,D7.1,&
             1X,1P,D24.16,1X,1P,D7.1,1X,1P,D24.16,1X,1P,D7.1,1X,1P,D7.1,&
             1X,1P,D7.1,1X,1P,D7.1,1X,I3,1X,I1,1X,L1,1X,I6,1X,I6,1X,I3, &
             1X,I7,1X,I7,1X,I2,1X,I7,1X,I7,1X,I7,0P,F8.2)

end subroutine algencan
